# coding: latin-1
# Description: Correlation Matrix Test
# Version: 1.5
# Author: R. Kessel
print 'Correlation Matrix Test (CorrMatrixTest.py Version: 1.5)'
print 'Author: Rdiger Kessel (r.kessel@metrodata.de)'
try:
    import Scripter
    GUMWB=Gumwb()
    GUMWBscript=True
except NameError:
    GUMWB = None
    GUMWBscript=False

def Msg(id):
    if GUMWBscript:
        return Scripter.MsgStr(id)
    else:
        return ""

import math
import numpy as N

Limit1 = -0.1 # limit of the eigenvalues for correction
Limit2 = 0.01  # limit of the norm of the correction

def CompSpec(EW,EV):
    '''
    Spectral decomposition of a correlation matrix from eigenvalues and eigenvectors
    Parameter: EW: array [n] of eigenvalues
               EV: array [n,n] of eigenvectors
    Return: Array[n,n] containing the correlation matrix
    The algorithm is based of the section Spectral decomposition in
        Riccardo Rebonato, Peter Jaeckel: The most general methodology to create a valid
        correlation matrix for risk management and option pricing purposes,
        http://www.quarchome.org/correlationmatrix.pdf
    '''
    L=N.matrix(N.zeros((len(EW),len(EW)),dtype='float64'))
    T=N.matrix(N.zeros((len(EW),len(EW)),dtype='float64'))
    for i in xrange(len(EW)):
        if EW[i]>=0:
            L[i,i]=math.sqrt(EW[i])
    for i in xrange(len(EW)):
        Ti=0
        for j in xrange(len(EW)):
            Ti+=EV[i,j]*EV[i,j]*EW[j]
        T[i,i]=1/math.sqrt(Ti)
    B=T*EV*L
    M=B*B.T
    for i in xrange(len(EW)):
        M[i,i]=1.0
    return N.array(M)

def LMN(m,M1,M2):
    '''
    Modified Least Maximum Norm
    Parameter: m: 0 - standard least maximum norm, 1 - modified least maximum norm
               M1, M2: array [n,n] or matrix [n,n] matrix to compare
    Return: float containing the least maximum norm
    LMN will calculate the larges component wise difference between M1 and M2
    '''
    n1,m1=M1.shape
    n2,m2=M2.shape
    if (n1<>n2) or (m1<>m2) or (n1<>m1):
        FATALERROR(254,"Internal Error: Matrix structure does not match!")
    max=0
    f=1
    for i in xrange(n1):
        for j in xrange(n1):
            if m==1:
                if M1[i,j]>=M2[i,j]:
                    f=math.fabs(M1[i,j])
                else:
                    f=math.fabs(M2[i,j])
                if f<0.1:
                    f=0.1
            d=math.fabs(M1[i,j]-M2[i,j])*f
            if d>max:
                max=d
    return max

def MatrixTest(DM,DC):
    '''
    Correlation Matrix Test
    Parameter: DM: list of list [[DC]..[DC]]
               DC: number columns and rows
    Return: Boolean indicates if the matrix is positive semidefinite

    The function will check if the given matrix DM is positive semidefinite

    Hint: The function is part of the interface to the CorrMatrixTest expert.
          The name and parameter must not be changed!
    '''
    print
    print 'Start correlation matrix test'
    if len(DM)==0:
        psd=True
    else:
        M=N.matrix(DM)
        isModi=False
        try:
            CM=N.linalg.cholesky(M)
            print 'Cholesky decomposition passed'
            psd=True
        except N.linalg.LinAlgError, e:
            print 'Cholesky decomposition failed'
            print 'Try eigenvalue decomposition'
            EW,EV=N.linalg.eigh(M)
            psd=True
            for W in EW:
                if W<0:
                    psd=False
            if psd:
                print 'All eigenvalues greater zero'
            if not psd:
                m=True
                for i,W in enumerate(EW):
                    if W<Limit1:
                        m=False
                    elif W<1E-10:
                        EW[i]=1E-10
                if not m:
                    print 'At least one eigenvalue is negative'
                if m:
                    MM=CompSpec(EW,EV)
                    isModi=True
                    try:
                        CM=N.linalg.cholesky(MM)
                        psd=True
                    except N.linalg.LinAlgError, e:
                        EW,EV=N.linalg.eigh(MM)
                        psd=True
                        for W in EW:
                            if W<0:
                                psd=False
                    if psd:
                        lmn=LMN(0,M,MM)
                        if lmn>Limit2:
                            print 'Negative eigenvalues could not be corrected!'
                            print 'Least Maximum Norm is: %1.3f' % lmn
                            psd=False
                        else:
                            print 'Negative eigenvalues could be corrected'
                            print 'Least Maximum Norm is: %1.3f' % lmn
                    else:
                        print 'Shift of eigenvalues failed!'
    if psd:
        s=Msg(200)
        print '--> Correlation matrix is positive semi-definite'
    else:
        s=Msg(201)
        print '--> Correlation matrix is not positive semi-definite'
    if GUMWBscript:
        Model=GUMWB.ActiveModel
        from datetime import date
        Desc=Model.InpCorrelation.Description
        if (Desc.Text==None) or (Desc.Text==''):
            lf=''
        else:
            lf='\r\n'
        if s.find('%s')>=0:
            s=s % str(date.today())
        Desc.Append(lf+s)
    print 'Done.'
    return psd

def Decomposition(DM,DC):
    '''
    Eigenvalue decomposition
    Parameter: DM: list of list [[DC]..[DC]]
               DC: number columns and rows
    Return: list of arrays containing the eigenvalues and eigenvectors

    The function will calculate the eigenvalues and eigenvectors of the given
    correlation matrix

    Hint: The function is part of the interface to the CorrMatrixTest expert.
          The name and parameter must not be changed!
    '''
    print
    print 'Start eigenvalue decomposition'
    if len(DM)==0:
        R=[[],[]]
    else:
        M=N.matrix(DM)
        EW,EV=N.linalg.eigh(M)
        R=[N.array(EW),N.array(EV)]
    print 'Done.'
    return R




